#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
# Creation Date : 2019-05-02 - 14:09:56
#
# pylint: disable=too-many-locals
# pylint: disable=too-many-instance-attributes
# pylint: disable=no-member
"""
-----------
Graphic utilities for nsfds2

@author: Cyril Desjouy
"""

import os
import sys
import getpass
import h5py
import numpy as _np
import matplotlib.pyplot as _plt
import matplotlib.animation as _ani
from mpl_toolkits.axes_grid1 import make_axes_locatable
from ofdlib2 import fdtd as _fdtd, derivation as _derivation
from progressbar import ProgressBar, Bar, ReverseBar, ETA
from mplutils import modified_jet, MidpointNormalize, set_figsize, get_subplot_shape
import fdgrid.graphics as _graphics


__all__ = ['get_data', 'DataIterator', 'DataExtractor', 'Plot']


def get_data(filename):
    """ Get data from filename. """

    try:
        home = os.path.expanduser("~")
        filename = filename.replace('~', home)
        data = h5py.File(filename, 'r')
    except OSError:
        print('You must provide a valid hdf5 file')
        sys.exit(1)
    else:
        return data


class DataIterator:
    """ Data Generator

    Parameters
    ----------

    data : DataExtractor instance
    view : The variable to display. string.
    nt : The last frame to consider. int

    """

    def __init__(self, data_extractor, view=('p'), nt=None):

        self.data = data_extractor
        self.view = view
        self.ns = self.data.get_attr('ns')
        self.icur = -self.ns
        if nt is None:
            self.nt = self.data.get_attr('nt')
        else:
            self.nt = nt

    def __iter__(self):
        """ Iterator """

        return self

    def __next__(self):
        """ Next element of iterator : (frame_number, variable) """

        try:
            self.icur += self.ns
            if self.nt:
                if self.icur > self.nt:
                    raise StopIteration

            tmp = [self.icur]
            for var in self.view:
                tmp.append(self.data.get(view=var, iteration=self.icur))

            return tmp

        except KeyError:
            raise StopIteration


class DataExtractor:
    """ Extract data from hdf5 file

    Parameters
    ----------

    data : Path to hdf5 file (str) or data from hdf5 file

    """

    def __init__(self, data):

        if isinstance(data, str):
            self.data = get_data(data)
        else:
            self.data = data

        self.var = {'p':'p', 'rho':'rho', 'vx':'rhou', 'vz':'rhov', 'e':'rhoe'}
        self.ns = self.get_attr('ns')

        if self.get_attr('mesh') == 'curvilinear':
            self.J = self.get_dataset('J')
        else:
            self.J = _np.ones((self.get_attr('nx'), self.get_attr('nz')))

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def reference(self, view='p', ref=None):
        """ Generate the references  for min/max colormap values """

        if ref is None:
            ref = self.autoref(view=view)

        if isinstance(ref, int):
            var = self.get(view=view, iteration=ref)
            return var.min(), var.max()

        if isinstance(ref, tuple):
            varmin = self.get(view=view, iteration=ref[0])
            varmax = self.get(view=view, iteration=ref[1])
            return varmin.min(), varmax.max()

        print("Only 'p', 'rho', 'vx', 'vz', 'vort' and 'e' available !")
        sys.exit(1)

    def autoref(self, view='p'):
        """ Autoset reference. """

        var = DataIterator(self, view=(view, ))

        maxs = []
        mins = []
        for _, v in var:
            maxs.append(v.max())
            mins.append(v.min())

        maxs = _np.array(maxs)
        mins = _np.array(mins)

        refmax = abs(maxs - maxs.mean()).argmin()*self.ns
        refmin = abs(mins - mins.mean()).argmin()*self.ns

        return refmin, refmax

    def close(self):
        """ Close hdf5 file """
        self.data.close()

    def get(self, view='p', iteration=0):
        """ Get data at iteration. """

        if view == 'p':
            rho = self.data[f"rho_it{iteration}"][...]*self.J
            rhou = self.data[f"rhou_it{iteration}"][...]*self.J
            rhov = self.data[f"rhov_it{iteration}"][...]*self.J
            rhoe = self.data[f"rhoe_it{iteration}"][...]*self.J
            p = _np.empty_like(rho)
            _fdtd.p(p, rho, rhou, rhov, rhoe, self.data.attrs['gamma'])

            return p.T - self.data.attrs['p0']

        if view == 'rho':
            return (self.data[f"{view}_it{iteration}"][...]*self.J).T

        if view in ['vx', 'vz', 'e']:
            vx = (self.data[f"{self.var[view]}_it{iteration}"][...]*self.J)
            rho = (self.data[f"rho_it{iteration}"][...]*self.J)
            return (vx/rho).T

        if view == 'vort':
            rho = (self.data[f"rho_it{iteration}"][...]*self.J)
            vx = (self.data[f"rhou_it{iteration}"][...]*self.J)/rho
            vz = (self.data[f"rhov_it{iteration}"][...]*self.J)/rho
            return _derivation.vorticity(self.data['x'][...], self.data['x'][...],
                                         vx, vz).T

        raise ValueError("view must be 'p', 'rho', 'vx', 'vz', 'e', or 'vort'")

    def get_attr(self, attr):
        """ Get attribute from hdf5 file. attr must be string."""
        return self.data.attrs[attr]

    def get_dataset(self, dataset):
        """ Get dataset from hdf5 file. attr must be string."""

        return self.data[dataset][...]


class Plot:
    """ Helper class to plot results from nsfds2.

    Parameters
    ----------

    filename : hdf5 file
    quiet : quiet mode. Boolean.
    """

    def __init__(self, filename, quiet=False):

        self.filename = filename
        self.path = os.path.dirname(filename) + '/'
        self.quiet = quiet

        self.data = DataExtractor(self.filename)
        self.nt = self.data.get_attr('nt')

        self._init_geo()
        self._init_fig()

    def _init_geo(self):
        """ Init coordinate system. """

        self.obstacles = self.data.get_attr('obstacles')
        self.Npml = self.data.get_attr('Npml')
        self.mesh = self.data.get_attr('mesh')
        self.bc = self.data.get_attr('bc')

        if self.mesh == 'curvilinear':
            self.x = self.data.get_dataset('xp')
            self.z = self.data.get_dataset('zp')
        else:
            self.x, self.z = _np.meshgrid(self.data.get_dataset('x'), self.data.get_dataset('z'))
            self.x = _np.ascontiguousarray(self.x.T)
            self.z = _np.ascontiguousarray(self.z.T)

    def _init_fig(self):
        """ Init figure parameters. """

        self.cm = modified_jet()
        self.title = r'{} -- iteration : {}'
        self.titles = {'p': r'$p_a$ [Pa]',
                       'e': r'$e$ [kg.m$^2$.s$^{-2}$]',
                       'vx': r'$v_x$ [m/s]',
                       'vz': r'$v_z$ [m/s]',
                       'rho': r'$\rho$ [kg.m$^3$]',
                       'vort': r'$\omega$ [m/s]'}

    def movie(self, view=('p', 'e', 'vx', 'vz'), nt=None, ref=None,
              figsize='auto', show_pml=False, show_probes=False, dpi=100):
        """ Make movie. """

        # Progress bar
        if not self.quiet:
            widgets = [Bar('>'), ' ', ETA(), ' ', ReverseBar('<')]
            pbar = ProgressBar(widgets=widgets, maxval=self.nt).start()

        # Movie parameters
        title = os.path.basename(self.filename).split('.')[0]
        metadata = dict(title=title, artist=getpass.getuser(), comment='From nsfds2')
        writer = _ani.FFMpegWriter(fps=24, metadata=metadata, bitrate=-1, codec="libx264")
        movie_filename = f'{title}.mkv'

        # Nb of iterations
        if nt is None:
            nt = self.nt

        # Create Iterator and make 1st frame
        data = DataIterator(self.data, view=view, nt=nt)
        i, *var = next(data)
        fig, axes, ims = self.fields(view=view, iteration=i, ref=ref,
                                     show_pml=show_pml,
                                     show_probes=show_probes,
                                     figsize=figsize)

        with writer.saving(fig, self.path + movie_filename, dpi=dpi):

            writer.grab_frame()

            for i, *var in data:

                # StackOv : using-set-array-with-pyplot-pcolormesh-ruins-figure
                for ax, mesh, v, j in zip(axes.ravel(), ims, var, range(len(ims))):
                    mesh.set_array(v[:-1, :-1].T.flatten())
                    ax.set_title(self.titles[view[j]] + f' (n={i})')

                writer.grab_frame()

                if not self.quiet:
                    pbar.update(i)

            if not self.quiet:
                pbar.finish()

    def probes(self):
        """ Plot probes. """

        probes = self.data.get_dataset('probes_location').tolist()

        if not probes:
            return None

        p = self.data.get_dataset('probes_value')
        t = _np.arange(self.nt)*self.data.get_attr('dt')

        _, ax = _plt.subplots(figsize=(9, 4))
        for i, c in enumerate(probes):
            if self.data.get_attr('mesh') == 'curvilinear':
                p0 = self.data.get_attr('p0')/self.data.get_dataset('J')[c[0], c[1]]
            else:
                p0 = self.data.get_attr('p0')
            ax.plot(t, p[i, :] - p0, label=f'@{tuple(c)}')
        ax.set_xlim(t.min(), t.max())
        ax.set_xlabel('Time [s]')
        ax.set_ylabel('Pressure [Pa]')
        ax.legend()
        ax.grid()

        return None

    def fields(self, view=('p', 'e', 'vx', 'vz'), iteration=None, ref=None,
               show_pml=False, show_probes=True, figsize='auto'):
        """ Make figure """

        if iteration is None:
            iteration = self.nt

        var = []
        norm = []
        ims = []

        for v in view:
            var.append(self.data.get(view=v, iteration=iteration).T)
            vmin, vmax = self.data.reference(view=v, ref=ref)
            norm.append(MidpointNormalize(vmin=vmin, vmax=vmax, midpoint=0))

        fig, axes = _plt.subplots(*get_subplot_shape(len(var)))

        if not isinstance(axes, _np.ndarray):   # if only 1 varible in view
            axes = _np.array(axes)

        for i, ax in enumerate(axes.ravel()):
            if i < len(var):
                ims.append(ax.pcolormesh(self.x, self.z, var[i],
                                         cmap=self.cm, norm=norm[i]))
                ax.set_title(self.titles[view[i]] + f' (n={iteration})')
                ax.set_xlabel(r'$x$ [m]')
                ax.set_ylabel(r'$z$ [m]')
                ax.set_aspect('equal')
                divider = make_axes_locatable(ax)
                cax = divider.append_axes("right", size="5%", pad=0.05)
                _plt.colorbar(ims[i], cax=cax)

                probes = self.data.get_dataset('probes_location').tolist()
                if probes and show_probes:
                    _ = [ax.plot(self.x[i, j], self.z[i, j], 'ro') for i, j in probes]

                _graphics.plot_subdomains(ax, self.x, self.z, self.obstacles)
                if show_pml:
                    _graphics.plot_pml(ax, self.x, self.z, self.bc, self.Npml)
            else:
                ax.remove()

        fig.set_size_inches(*set_figsize(axes, figsize))
        _plt.tight_layout()

        return fig, axes, ims

    @staticmethod
    def show():
        """ Show all figures. """
        _plt.show()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : 2019-02-13 - 10:58:09
"""
-----------

Examples of obstacle arangements.

@author: Cyril Desjouy
"""


import numpy as _np
from fdgrid import Domain, Obstacle


def logo(xn, zn):
    """ Curvlinear coordinates : test case 1. Physical == numerical """

    xp = xn.copy()
    zp = zn \
        + _np.linspace(0.5, 0, zn.shape[1])*(100*xp**2)
    return xp, zp


def letter_a(nx, nz, size=10):
    """ letter a.

    Parameters:
    ----------

    size : percentage of domain
    """

    size = size/100
    xc, zc = int(nx/2), int(nz/2)
    xmin, xmax = int(xc-size*nx), int(xc+size*nx)
    zmin, zmax = int(zc-2*size*nz), int(zc+2*size*nz)

    geo = [Obstacle([xmin, zc-6, xmax, zc+6], 'RRRR'),
           Obstacle([xmin, zmax-12, xmax, zmax], 'RRRR'),
           Obstacle([xmin-12, zmin, xmin, zmax], 'RRRR'),
           Obstacle([xmax, zmin, xmax+12, zmax], 'RRRR')]

    return Domain((nx, nz), data=geo)


def moving_square(nx, nz, size_percent=20):
    """ Square in the middle.

    Parameters:
    -----------

    size_percent (float): size of the square in percent of the largest
    dimension of the domain.
    """

    size = int(min(nx, nz)*size_percent/100)
    obs1 = Obstacle([int(nx/2)-size, int(nz/2)-size,
                     int(nx/2)+size, int(nz/2)+size], 'URRR')
    obs2 = Obstacle([nx-11, 0, nx-1, nz-1], 'URRR')

    obs1.set_moving_bc({'f': 70000, 'A': 1, 'func': 'sine'})
    obs2.set_moving_bc({'f': 73000, 'A': 1, 'func': 'sine'})

    return Domain((nx, nz), data=[obs1, obs2])




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
# Creation Date : 2019-03-21 - 23:43:11
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import os
import sys
from fdgrid import templates as _tplt


def get_obstacle(cfg):
    """ Get obstacle from custom file or fdgrid templates. """

    if cfg.geofile != 'None':
        try:
            sys.path.append(os.path.dirname(cfg.geofile))
            custom = __import__(os.path.basename(cfg.geofile).split('.')[0])
            obstacle = getattr(custom, cfg.geoname)(cfg.nx, cfg.nz)
            cfg.geoflag = True
        except (AttributeError, ImportError):
            cfg.geoflag = False
            obstacle = []
    else:
        try:
            obstacle = getattr(_tplt, cfg.geoname)(cfg.nx, cfg.nz)
            cfg.geoflag = True
        except AttributeError:
            cfg.geoflag = False
            obstacle = []
    return obstacle


def get_curvilinear(cfg):
    """ Get curvilinear fonction from custom file or fdgrid templates. """

    if cfg.geofile != 'None':
        try:
            sys.path.append(os.path.dirname(cfg.geofile))
            custom = __import__(os.path.basename(cfg.geofile).split('.')[0])
            fcurv = getattr(custom, cfg.curvname)
            cfg.curvflag = True
        except (AttributeError, ImportError):
            cfg.curvflag = False
            fcurv = None
    else:
        try:
            fcurv = getattr(_tplt, cfg.curvname)
            cfg.curvflag = True
        except AttributeError:
            cfg.curvflag = False
            fcurv = None
    return fcurv



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
# Creation Date : 2019-06-21 - 12:30:57
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import os as _os
import numpy as _np
import scipy.signal as _sps
import scipy.io.wavfile as _wf
from nsfds2.utils.misc import colors as _colors
from nsfds2.utils import DataExtractor as _DataExtractor

def resample(file, target_rate, pad=None, write=False, force_mono=True):
    """ Resample target wave file with target_rate. """

    target_rate = int(target_rate)

    if not 1 < target_rate < 4.3e6:
        raise ValueError('Sampling rate must be 1 < rate < 4.3e6')

    path = _os.path.dirname(file)
    filename = _os.path.basename(file).split('.')
    rate, data = _wf.read(file)
    dtype = data.dtype
    duration = data.shape[0]/rate
    N = int(target_rate*duration)

    if len(data.shape) == 2 and force_mono:   # stereo to mono
        data = (data[:, 0] + data[:, 1])/2

    print(_colors.BLUE + f'Resampling {file} at {target_rate} kHz ({N} points)...')
    print(_colors.RED + f'Set nt > {N} to play the whole sound' + _colors.END)
    if len(data.shape) == 1:   # mono
        data_r = _sps.resample(data, N).astype(dtype)
        if pad:
            data_r = get_padded(data_r, pad)


    if len(data.shape) == 2:   # stereo
        tmp_l = _sps.resample(data[:, 0], N).astype(dtype)
        tmp_r = _sps.resample(data[:, 1], N).astype(dtype)
        if pad:
            tmp_l = get_padded(tmp_l, pad)
            tmp_r = get_padded(tmp_r, pad)

        data_r = _np.vstack([tmp_l, tmp_r]).T


    if write:
        print(f'Writing {N} samples at {target_rate} kHz rate...')
        _wf.write(path + '{}_r.{}'.format(*filename), rate=target_rate, data=data_r)

    return data_r/abs(data_r).max()


def get_padded(s, N, value=0):
    """ Pad signal with value. """
    if N > s.shape[0]:
        return _np.concatenate([s, value*_np.ones(N - s.shape[0])])

    return s


def normalize(s, dtype=_np.int16):
    """ Normalize wav data. """
    smax = _np.iinfo(dtype).max
    return (s/abs(s).max()*smax).astype(dtype)


def probes_to_wave(datafile, dtype=_np.int16, path=None):
    """ Make wav files from probes.

    Parameters
    ----------

    datafile: the hdf5 file
    dtype: type of the output wave file (np.int8, np.int16)
    path: path to save wavefiles

    Returns
    -------

    out: list of probe signals saved in the wave files
    """

    probes = []
    home = _os.path.expanduser('~')
    datafile = datafile.replace('~', home)

    with _DataExtractor(datafile) as data:
        probes_values = data.get_dataset('probes_value')
        p0 = data.get_attr('p0')
        dt = data.get_attr('dt')

    if not path:
        path = _os.path.dirname(datafile)


    if list(probes_values):

        filename = _os.path.basename(datafile).split('.')[0]

        for i, p in enumerate(probes_values):
            name = f'{path}/{filename}_probe_{i}.wav'
            tmp = normalize(p - p0, dtype=dtype)
            print(f'Writing {name}...')
            _wf.write(name, rate=int(1/dt), data=tmp)
            probes.append(tmp)
        print('Done!')

    else:
        print('No probes found!')

    return probes



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : 2018-04-13 00:19:03
"""
-----------

The SelectiveFilter class propose a filtering procedure to reduce Gibbs
oscillations.

@author: Cyril Desjouy
"""

import ofdlib2.filters as flt

class SelectiveFilter:
    """ Filter rho, rhou, rhov and rhoe. """


    def __init__(self, msh, fld, cfg):

        self.msh = msh
        self.fld = fld
        self.flt = flt.sfilter(msh.nx, msh.nz, cfg.xnu, stencil=cfg.flt_stencil)

        for sub in self.msh.fmdomains:
            bc = sub.bc.replace('.', '').replace('U', 'R').replace('V', 'R').replace('W', 'R')
            sub.filt_method = getattr(self.flt, f"f{sub.axname}_{bc}")

    def apply(self):
        """ Dispatch filtering. """

        for sub in self.msh.fxdomains:
            sub.filt_method(self.fld.r, self.fld.K, *sub.ix, *sub.iz)
            sub.filt_method(self.fld.ru, self.fld.Ku, *sub.ix, *sub.iz)
            sub.filt_method(self.fld.rv, self.fld.Kv, *sub.ix, *sub.iz)
            sub.filt_method(self.fld.re, self.fld.Ke, *sub.ix, *sub.iz)

        self.update(self.msh.fxdomains)

        for sub in self.msh.fzdomains:
            sub.filt_method(self.fld.r, self.fld.K, *sub.ix, *sub.iz)
            sub.filt_method(self.fld.ru, self.fld.Ku, *sub.ix, *sub.iz)
            sub.filt_method(self.fld.rv, self.fld.Kv, *sub.ix, *sub.iz)
            sub.filt_method(self.fld.re, self.fld.Ke, *sub.ix, *sub.iz)

        self.update(self.msh.fzdomains)

    def update(self, domains):
        """ Update fields. """
        for sub in domains:
            self.flt.apply(self.fld.r, self.fld.K, *sub.ix, *sub.iz)
            self.flt.apply(self.fld.ru, self.fld.Ku, *sub.ix, *sub.iz)
            self.flt.apply(self.fld.rv, self.fld.Kv, *sub.ix, *sub.iz)
            self.flt.apply(self.fld.re, self.fld.Ke, *sub.ix, *sub.iz)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : mar. 10 avril 2018 17:52:42 CEST
# Last Modified : ven. 11 mai 2018 16:13:55 CEST
"""
-----------

setup file for nsfds2

@author: Cyril Desjouy
"""

from setuptools import setup, find_packages

setup(
    name='nsfds2',
    description="Finite difference solver for Navier-Stokes equations",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    version="0.9.5",
    license="GPL",
    url='https://github.com/ipselium/nsfds2',
    author="Cyril Desjouy",
    author_email="cyril.desjouy@univ-lemans.fr",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["numpy", "matplotlib", "ofdlib2>=0.9.5",
                      "progressbar33", "mplutils>=0.3.0", "h5py", "fdgrid>=0.8.4"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
    entry_points={
        'console_scripts': [
            'nsfds2 = nsfds2.solver:main',
        ],
    }
)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
# Creation Date : 2019-03-08 - 10:47:50
#
# pylint: disable=redefined-builtin
"""
-----------

Headers for nsfds2

@author: Cyril Desjouy
"""

import os
import sys
import platform
import textwrap
import numpy
import ofdlib2
import fdgrid
import nsfds2


def _columns():
    try:
        _, col = os.popen('stty size', 'r').read().split()
        return int(col) if int(col) < 81 else 80
    except ValueError:
        return 80


def copyright():
    """ Show copyright. """
    col = 76 if _columns() > 76 else _columns
    cp = [col*'#',
          "nsfds2 v{} -- Copyright (C) 2016-2019 -- Cyril Desjouy".format(nsfds2.__version__),
          " "]
    lc = "This program comes with ABSOLUTELY NO WARRANTY. " + \
         "This is free software, and you are welcome to redistribute it " + \
         "under certain conditions; See GNU GPL v3 for more details."
    cp += textwrap.wrap(lc, int(col))
    cp += [col*"#"]
    tmp = ''
    for txt in cp:
        tmp += "# {:^{col}} #\n".format(txt, col=col)
    print(tmp)


def version():
    """ Display versions of modules in use. """

    print('System : {}\n'.format(platform.platform()))
    print('\tpython : {:<10}\tnumpy  : {:<10}'.format(sys.version.split(' ')[0], numpy.__version__))
    print('\tfdgrid : {:<10}\tofdlib : {:<10}'.format(fdgrid.__version__, ofdlib2.__version__))
    print('\n' + _columns()*"#" + "\n")


def start(cfg):
    """ Prompt start. """

    check_geo(cfg)

    log = "Starting computation for geometry : '{}' ({}x{})."
    print(log.format(cfg.geoname, cfg.nx, cfg.nz))
    key = input("Hit enter to continue (prefix 'q' to abort) ! ")
    if key == 'q':
        sys.exit(0)


def check_geo(cfg):
    """ Check geofile. """

    if not cfg.geoflag:
        print('Warning: Unable to load {} from {}'.format(cfg.geoname, cfg.geofile))
        cfg.geoname = 'empty'


def parameters(cfg, msh):
    """ Show simulation parameters. """

    s = msh.__str__()
    if 'A' in cfg.bc:
        nsfds2.init.fields.set_sigma(cfg, msh)
        s += "\t* sigma=({:.3e}, {:.3e}) and beta={}\n".format(cfg.sigmax,
                                                               cfg.sigmaz,
                                                               cfg.beta)

    s += "\t* geometry : '{}'\n".format(cfg.geoname)

    s += "\t* dt = {:.5e} s and nt = {}.\n".format(cfg.dt, cfg.nt)

    if cfg.stype not in cfg.none:
        s += "\t* source : {} at ({}, {})".format(cfg.stype, cfg.ixS, cfg.izS)
        if cfg.ftype == 'harmonic':
            s += 'with f={}.\n'.format(cfg.f0)
        else:
            s += '.\n'

    if cfg.ftype not in cfg.none:
        s += "\t* flow   : {} (U0={}, V0={}).".format(cfg.ftype, cfg.U0, cfg.V0)

    print(s)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : 2019-03-01 - 10:29:55
#
# pylint: disable=too-few-public-methods
"""
-----------

Compute Derivatives

@author: Cyril Desjouy
"""


import ofdlib2.derivation as drv


class Cin:
    """ Dispatch domains to associated computation functions. """

    def __init__(self, msh, fld, cfg):

        self.msh = msh
        self.fld = fld
        self.cfg = cfg
        self.du = drv.du(msh.x, msh.z, msh.stencil)

        for sub in self.msh.dmdomains:
            bc = sub.bc.replace('.', '').replace('U', 'R')
            bc = bc.replace('V', 'R').replace('W', 'R')
            sub.cin_method = getattr(self.du, f'dud{sub.axname}_{bc}')

    def dispatch(self):
        """ Dispatch the domains to the functions. """

        # dE/dz ---------------------------------------------------------------
        if self.cfg.mesh in ['regular', 'adaptative']:
            self.fld.fdtools.Eu(self.fld.E, self.fld.Eu, self.fld.Ev, self.fld.Ee,
                                self.fld.r, self.fld.ru, self.fld.rv, self.fld.re,
                                self.fld.p)

        elif self.cfg.mesh == 'curvilinear':
            self.fld.fdtools.EuJ(self.fld.E, self.fld.Eu, self.fld.Ev, self.fld.Ee,
                                 self.fld.r, self.fld.ru, self.fld.rv, self.fld.re,
                                 self.fld.p, self.msh.dxn_dxp, self.msh.dxn_dzp)

        for sub in self.msh.dxdomains:
            sub.cin_method(self.fld.E, self.fld.K, *sub.ix, *sub.iz)
            sub.cin_method(self.fld.Eu, self.fld.Ku, *sub.ix, *sub.iz)
            sub.cin_method(self.fld.Ev, self.fld.Kv, *sub.ix, *sub.iz)
            sub.cin_method(self.fld.Ee, self.fld.Ke, *sub.ix, *sub.iz)

        # dF/dz ---------------------------------------------------------------
        if self.cfg.mesh in ['regular', 'adaptative']:
            self.fld.fdtools.Fu(self.fld.F, self.fld.Fu, self.fld.Fv, self.fld.Fe,
                                self.fld.r, self.fld.ru, self.fld.rv, self.fld.re,
                                self.fld.p)

        elif self.cfg.mesh == 'curvilinear':
            self.fld.fdtools.FuJ(self.fld.F, self.fld.Fu, self.fld.Fv, self.fld.Fe,
                                 self.fld.r, self.fld.ru, self.fld.rv, self.fld.re,
                                 self.fld.p, self.msh.dzn_dxp, self.msh.dzn_dzp)

        for sub in self.msh.dzdomains:
            sub.cin_method(self.fld.F, self.fld.K, *sub.ix, *sub.iz)
            sub.cin_method(self.fld.Fu, self.fld.Ku, *sub.ix, *sub.iz)
            sub.cin_method(self.fld.Fv, self.fld.Kv, *sub.ix, *sub.iz)
            sub.cin_method(self.fld.Fe, self.fld.Ke, *sub.ix, *sub.iz)
